require 'uri'
require 'securerandom'
require 'uu/os/application/abstract_controller'
require 'uu/os/application/immediate_response'
require 'uu/os/application/config'
require 'uu/os/cmd/execution_context'
require 'uu/os/cmd/callback_execution_context'
require 'uu/os/http/http_request'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_codes'
require 'uu/os/ipc/error_kind'
require 'uu/os/application/application_error'
require 'uu/os/application/error_codes'
require 'uu/os/security/session'
require 'uu/os/rest/client_exception'
require 'uu/os/rest/remote_exception'
require 'uu/os/logger'

module UU
  module OS
    module CMD

      # Ancestor for development of custom command controllers.
      # All controllers must define +PATH+ constant (for controller
      # mapping) and should define at least {#on_exec on_exec} method.
      # Additionally, it is possible to modify default processing
      # behavior by overriding methods declared in
      # {UU::OS::Application::AbstractController AbstractController}.
      class CommandController < UU::OS::Application::AbstractController

        # Main action method of controller.
        # @param [UU::OS::CMD::ExecutionContext] ctx Instance of execution
        #    context.
        # @return [Object] Result of command (any object to be serialized
        #    via +to_json+ method, or response provided by {UU::OS::Server::AbstractServlet#response
        #    response}, {UU::OS::Server::AbstractServlet#file_download_response file_download_response}
        #    or {UU::OS::Server::AbstractServlet#streamed_response streamed_response} methods).
        def on_exec(ctx)
          raise UU::OS::CMD::CommandError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::IPC::ErrorCodes::NOT_IMPLEMENTED,
            "Method on_exec is not implemented on #{self.class.name}.")
        end

        # Checks if invocation method (POST, GET, etc.) matches the
        # controller type. By default, "get*" commands should be
        # invoked as GET, any other commands should be invoked as POST.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [void]
        # @raise [UU::OS::Application::ControllerError] In case invocation
        #    method does not match the expected one.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        def assert_invocation_method(request)
          parts = request.url.split('/')
          if parts[-1] =~ /^get/ || parts[-2] =~ /^get/
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
              "Invalid invocation method, expected GET or POST.") if request.method != :get && request.method != :post
          else
            raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::INVALID_REQUEST,
              "Invalid invocation method, expected POST.") if request.method != :post
          end
        end

        # Checks if binary data upload is expected. By default binary
        # upload is expected for "*Data" commands invoked via POST
        # method.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [TrueClass, FalseClass] True if binary upload is expected.
        # @note Not intended to be used from controller implementation but for
        #    optional overriding of default behavior by custom implementation!
        def binary_expected?(request)
          parts = request.url.split('/')
          return request.method == :post && (parts[-1] =~ /.*Data$/ || parts[-2] =~ /.*Data$/)
        end

        # Initializes authn environment and evaluates incoming
        # authn token.
        # @private
        def prepare_execution(request, uu_uri, parameters) # :nodoc:
          super(request, uu_uri, parameters)
          if !request.headers.contains(:authorization)
            if request.method == :get
              realm = request.url.split('/')[2]
              header = {www_authenticate: "Basic realm=#{realm}"}
              raise UU::OS::Application::ImmediateResponse.new(401, header, "Missing authentication data.")
            else
              raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::NOT_AUTHENTICATED, "Missing authentication data.")
            end
          else
            begin
              UU::OS::Security::Session.login(request.headers[:authorization][0])
            rescue UU::OS::REST::ClientException => e
              if request.method == :get
                realm = request.url.split('/')[2]
                header = {www_authenticate: "Basic realm=#{realm}"}
                raise UU::OS::Application::ImmediateResponse.new(401, header, "Authentication failed.")
              else
                raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::CLIENT, UU::OS::Application::ErrorCodes::NOT_AUTHENTICATED, "Authentication failed.", e)
              end
            rescue Exception => e
              raise UU::OS::Application::ApplicationError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::Application::ErrorCodes::NOT_AUTHENTICATED, "Authentication failed.", e)
            end
          end
          # Prepare MDC context for logger.
          UU::OS::Logger::MDCProvider.put('uu-appURI', UU::OS::Application::Config[:app_uri]) unless UU::OS::Application::Config[:app_uri].nil?
          UU::OS::Logger::MDCProvider.put('uu-applogURI', UU::OS::Application::Config[:applog_uri]) unless UU::OS::Application::Config[:applog_uri].nil?
          UU::OS::Logger::MDCProvider.put('uu-ucURI', URI(request.url).path)
          UU::OS::Logger::MDCProvider.put('uu-execID', SecureRandom.uuid)
          UU::OS::Logger::MDCProvider.put('uu-entityURI', uu_uri)
          # TODO Solve setting of executive and access role to MDC
          # Executive role cannot be obtained without authorization
          # Reading of access role means additional performance hit
          # UU::OS::Logger::MDCProvider.put('uu-erURI', nil)
          # UU::OS::Logger::MDCProvider.put('uu-acrURI', UU::OS::Security::Session.get_access_role(uu_uri))
        end

        # Cleans up authn environment.
        # @private
        def finalize_execution(request, uu_uri, parameters) # :nodoc:
          super(request, uu_uri, parameters)
          UU::OS::Security::Session.logout(true)
          # Cleanup MCD logger context
          UU::OS::Logger::MDCProvider.get_context.each_key do |key|
            UU::OS::Logger::MDCProvider.remove(key)
          end
        end

        # Add command_v1 RemoteException to declared errors.
        # @private
        def declared_errors # :nodoc:
          dec_err = super.dup
          dec_err << UU::OS::REST::RemoteException
          return dec_err
        end

        # Returns context with injected session object (which is not
        # available on AbstractController level).
        # @private
        def get_context(request, uu_uri, parameters) # :nodoc:
          data = {}
          data[:request] = request
          data[:uu_uri] = uu_uri
          data[:parameters] = parameters
          data[:session] = UU::OS::Security::Session.current_session
          data[:forbid_async] = true
          ctx = parameters.delete(context_parameter_name)
          if !ctx.nil? && ctx.has_key?(:result)
            data[:result] = ctx[:result]
            return UU::OS::CMD::CallbackExecutionContext.new(data)
          else
            return UU::OS::CMD::ExecutionContext.new(data)
          end
        end

        # Translate co command error if necessary.
        # @private
        def handle_error(exception) # :nodoc:
          if !exception.kind_of?(UU::OS::IPC::IPCError) && !exception.kind_of?(UU::OS::REST::RemoteException)
            # Wrap undeclared error to common error.
            if UU::OS::Server::ServerContext.development_mode?
              wrapping_err = UU::OS::CMD::CommandError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::CMD::ErrorCodes::UNEXPECTED_PROBLEM, 'Unexpected server error.', exception)
            else
              # Do not publish internal problem when not in development mode.
              wrapping_err = UU::OS::CMD::CommandError.new(UU::OS::IPC::ErrorKind::SERVER, UU::OS::CMD::ErrorCodes::UNEXPECTED_PROBLEM, 'Unexpected server error.')
            end
            wrapping_err.instance_variable_set(:@id, exception.id)
            return wrapping_err
          else
            return exception  
          end
        end

        # Logs using {UU::OS::Logger}.
        # @private
        def log_error(context, message = nil, exception = nil) # :nodoc:
          message, exception = exception, message if message.kind_of?Exception
          @log = UU::OS::Logger.new(self.class)
          @log.error(message, exception)
        end

      end

    end
  end
end
