require 'timeout'
require 'securerandom'
require 'uu/os/application/abstract_controller'
require 'uu/os/cmd/command_executor'
require 'uu/os/security/session'
require 'uu/os/ipc/ipc_error'
require 'uu/os/ipc/error_kind'

module UU
  module OS
    module Util

      # Service providing system information of running module.
      # @private
      class ContainerService < UU::OS::Application::AbstractController # :nodoc:

        # Service path.
        # @private
        CONTROLLER_URL_PATH = '_uu_' # :nodoc:

        # Store reference to server context to be able to
        # provide various information.
        # @private
        def self.on_server_startup(ctx) # :nodoc:
          @@ctx = ctx
        end

        # Stores invocation time to thread.
        # @private
        def service(request) # :nodoc:
          Thread.current[:start] = Time.now
          super(request)
        end

        # Initializes authn environment and evaluates incoming authn token.
        # @private
        def prepare_execution(request, uu_uri, parameters) # :nodoc:
          super(request, uu_uri, parameters)
          if !request.headers.contains(:authorization)
            # Do not force authentication in development mode
            if !@@ctx.development_mode?
              if request.method =~ /^get$/i
                realm = request.url.split('/')[2]
                header = {www_authenticate: "Basic realm=#{realm}"}
                raise UU::OS::Application::ImmediateResponse.new(401, header, "Missing authentication data.")
              else
                raise UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Missing authentication data.")
              end
            end
          else
            begin
              UU::OS::Security::Session.login(request.headers[:authorization][0])
            rescue UU::OS::REST::ClientException => e
              if request.method =~ /^get$/i
                realm = request.url.split('/')[2]
                header = {www_authenticate: "Basic realm=#{realm}"}
                raise UU::OS::Application::ImmediateResponse.new(401, header, "Authentication failed.")
              else
                raise UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Authentication failed.", e)
              end
            rescue Exception => e
              raise UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::SERVER, nil, "Authentication failed.", e)
            end
          end
        end

        # Cleans up authn environment.
        # @private
        def finalize_execution(request, uu_uri, parameters) # :nodoc:
          super(request, uu_uri, parameters)
          UU::OS::Security::Session.logout(true)
        end

        # Returns health info.
        # @private
        def on_health(ctx) # :nodoc:
          start = Thread.current[:start]
          entity = ctx.uu_uri
          result = {reqId: SecureRandom.uuid()}
          result[:terrUri] = entity
          result[:queueUri] = nil
          result[:taskUri] = nil
          result[:trace] = []
          invocation = {cont: 'CMD'}
          invocation[:nodeId] = get_node_id(ctx)
          invocation[:timestamp] = start
          result[:trace] << invocation
          # If queue is defined, do asynchronous health
          queue = ctx.parameters[:queueUri]
          if !queue.nil? && !queue.empty?
            result[:queueUri] = queue
            # Prepare asynchronous invocation.
            acmd_uc_code = URI.split(ctx.request.url)[5]
            acmd_uc_code.gsub!(/(^\/)|(\/$)/, '')
            parts = acmd_uc_code.split('/')
            action = parts.delete_at(-1)
            app_url_code = parts.delete_at(0)
            cmd_url_path = parts.join('/')
            params = {}
            params[:queue_uri] = queue
            params[:app_url_code] = app_url_code
            params[:cmd_url_path] = cmd_url_path
            params[:action] = action
            params[:parameters] = {result: result}
            # Prepare callback
            callback = {}
            callback[:appUrlCode] = app_url_code
            callback[:cmdUrlPath] = cmd_url_path
            callback[:action] = :healthCallback
            callback[:mainObjectUri] = entity
            params[:callback] = callback
            # Invoke asynchronous command
            session = UU::OS::Security::Session.current_session
            client = UU::OS::CMD::CommandExecutor.new(session)
            begin
              task_uri = client.invoke(entity, params)
              result[:taskUri] = task_uri
            rescue Exception => e
              invocation[:exception] = e.to_json
            end
            if !task_uri.nil?
              # Wait for finish to append health result.
              timeout = ctx.parameters[:timeout] || 120
              health = get_health(client, task_uri, timeout)
              if health[0].kind_of?(String)
                invocation[:exception] = health[0]
              else
                result[:trace].push(*health[0])
              end
              if !health[1].nil?
                result[:trace].push(*health[1])
              end
            end
          end
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return result
        end

        # Asynchronous part of health.
        # @private
        def on_health_async(ctx) # :nodoc:
          start = Thread.current[:start]
          invocation = {cont: 'ASYNC_CMD'}
          invocation[:nodeId] = get_node_id(ctx)
          invocation[:timestamp] = start
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return invocation
        end

        # Callback health.
        # @private
        def on_health_callback(ctx) # :nodoc:
          start = Thread.current[:start]
          invocation = {cont: 'CALLBACK'}
          invocation[:nodeId] = get_node_id(ctx)
          invocation[:timestamp] = start
          invocation[:duration] = ((Time.now - start) * 1_000).to_i
          return invocation
        end

        # Returns server status.
        # @private
        def on_status(ctx) # :nodoc:
          if !@@ctx.development_mode?
            raise UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::CLIENT, nil, "Only available in development mode.")
          end
          status = {}
          if RUBY_ENGINE == 'jruby'
            status[:runtime] = "JRuby #{JRUBY_VERSION} (Ruby #{RUBY_VERSION}); Java #{java.lang.System.getProperty('java.version')} (#{java.lang.System.getProperty('java.vm.name')})"
          else
            status[:runtime] = "Ruby #{RUBY_VERSION}"
          end
          status[:uptime] = (Time.now - @@ctx.boot_time).to_i
          status[:libraries] = available_gems
          status[:configuration] = @@ctx.configuration
          status[:services] = @@ctx.registered_servlets
          return status
        end

        # Returns node ID.
        # @private
        def get_node_id(ctx) # :nodoc:
          node_id = ENV['UU_NODE_ID']
          if node_id.nil? && Object.const_defined?(:Java)
            node_id = Java::JavaLang::System.getProperty('uu.node.id')
          end
          if node_id.nil?
            node_id = ctx.request.headers[:host][0]
          end
          return node_id
        end

        # Returns hearbeat result of asynchronous invocation
        # (including health of callback).
        # @private
        def get_health(executor, task_uri, timeout) # :nodoc:
          begin
            Timeout::timeout(timeout.to_f) do
              state = executor.get_state(task_uri)
              while (state[:state] != 'FINISHED' && state[:state] != 'ERROR') ||
                (state[:callbackState] != 'FINISHED' && state[:callbackState] != 'ERROR')
                sleep(1)
                state = executor.get_state(task_uri)
              end
              result = []
              if state[:state] == 'ERROR'
                result << state[:exception]
                return result
              else
                result << state[:result]
              end
              if state[:callbackState] == 'FINISHED'
                result << state[:callbackResult]
              end
              return result
            end
          rescue Timeout::Error
            result = []
            # Try to get actual state, maybe everything just finished
            state = executor.get_state(task_uri)
            if state[:state] != 'ERROR' && state[:state] != 'FINISHED'
              # Asynchronous command was not yet invoked
              result << UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::SERVER, nil, 'Asynchronous command was not yet started.').to_json
              return result
            elsif state[:state] == 'ERROR'
              result << state[:exception]
              return result
            else
              result << state[:result]
            end
            if state[:callbackState] != 'ERROR' && state[:callbackState] != 'FINISHED'
              # Callback was not yet invoked, append exception to async container
              result[0][-2][:exception] = UU::OS::IPC::IPCError.new(UU::OS::IPC::ErrorKind::SERVER, nil, 'Callback was not yet started.').to_json
            elsif state[:callbackState] == 'FINISHED'
              result << state[:callbackResult]
            end
            return result
          end
        end

        # Returns list of available gems.
        # @private
        def available_gems # :nodoc:
          Gem::Specification.all = nil
          gems = Gem::Specification.map{|gem| "#{gem.name}:#{gem.version}"}
          Gem::Specification.reset
          gems
        end

      end

    end
  end
end
