require 'base64'
require 'set'
require 'uu/os/server/abstract_servlet'
require 'uu/os/server/parameter_parser'

module UU
  module OS
    module Util

      # Test implementation of session API used for command authentication.
      # This implementation is used only in case application server is started
      # in development mode and if it is explicitly allowed via environment
      # variable.
      # @private
      class EmbeddedAuthnServer < UU::OS::Server::AbstractServlet # :nodoc:
        include UU::OS::Server::ParameterParser

        HANDSHAKE_PATH = '/ues/wcp/ues/platform/security/authn/UESHandshakeService'

        SESSION_PATH = '/ues/wcp/ues/core/security/session/UESSession'

        @@accounts = Set.new

        @@tokens = Set.new

        # Register authentication services.
        # @private
        def self.on_server_startup(ctx) # :nodoc:
          accounts = ctx.configuration[:embedded_authn_accounts]
          # Use local authentication only in development mode and
          # if some testing accounts are defined.
          return if accounts.nil? || !ctx.development_mode?
          accounts.split(';').each do |account|
            @@accounts << account.strip if !account.nil?
          end
          ctx.register_servlet(self, true, HANDSHAKE_PATH)
          ctx.register_servlet(self, true, SESSION_PATH)
          base_url = Thread.current[:BASE_URL]
          if !base_url.nil?
            # Configure command client to invoke authentication locally.
            # (use legacy configuration parameters because session is invoked via command v1)
            handshake_cfg_key = 'uu.os.security.authn.ueshandshakeservice.server-url'
            session_cfg_key = 'uu.os.security.session.uessession.server-url'
            ctx.configuration.merge!({handshake_cfg_key => base_url, session_cfg_key => base_url})
            UU::OS::CMD::CommandClient.load_configuration({handshake_cfg_key => base_url, session_cfg_key => base_url})
          end
        end

        # Processes incoming request.
        # @private
        def service(request) # :nodoc:
          auth_token = request.headers[:authorization][0] if request.headers.contains(:authorization)
          return response(401, nil, 'Missing authentication token') if auth_token.nil? || auth_token.empty?
          method = request.url.split('/')[-1]
          if method =~ /handshake/
            if !eval_token(auth_token)
              return response(401, nil, 'Invalid authentication data')
            else
              return response(200, {content_type: 'application/json'}, 'true')
            end
          elsif method =~ /login/
            if !eval_token(auth_token)
              return response(401, nil, 'Invalid authentication data')
            else
              auth_token = "Bearer #{Base64.strict_encode64("#{auth_token}|#{Time.now}")}"
              @@tokens << auth_token
              return response(200, {content_type: 'application/json'}, "\"#{auth_token}\"")
            end
          else
            return response(404, nil, "Resource #{request.url} does not exist.")
          end
        end

        # Evaluate authentication token.
        # @private
        def eval_token(token) # :nodoc:
          if token =~ /^Basic (.*)/i
            return @@accounts.include?(Base64.strict_decode64($1))
          else
            return @@tokens.include?(token)
          end
        end

      end

    end
  end
end
